module MatchReg where

import AbsExtract
import Char
import ErrM
import Maybe (catMaybes)
import List (inits, tails)
import Text.Regex
import List

match :: Reg -> [(Ident,Either Reg Regex)] -> String -> Maybe [[(Ident,String)]]
match REps              env  []  = Just [[]]
match REps              env   _  = Nothing
match (RAlts str)   env [c]  = if (elem c str) then Just [[]] else Nothing
match (RAlts _)     env _    = Nothing
match (RAlt reg0 reg)   env str  =  
    case (match reg0 env str, match reg env str) of
     (Just xs, Just ys) -> case (xs,ys) of
                            ([[]],ys) -> return $ ys
			    (xs,[[]]) -> return $ xs
			    _         -> return $ xs ++ ys
     (Nothing, res)     -> res
     (res, Nothing)     -> res
match (RMinus reg0 reg) env str  =
  case(match reg0 env str, match reg env str) of
   (res, Nothing) -> res
   _              -> Nothing
match (RSeq reg0 reg)   env str  = 
   case (catMaybes 
      [aux (match reg0 env xs) (match reg env ys) | (xs,ys) <- split str]) of
    []   -> Nothing
    xs   -> Just $ concat xs
 where aux (Just xss) (Just yss) = 
	         return $ [xs ++ ys | xs <- xss, ys <- yss]
       aux _         _         = Nothing
match (RStar reg)       env []   = Just [[]]
match (RStar reg)       env str  = match (RSeq reg (RStar reg)) env str
match (RPlus reg)       env str  = match (RSeq reg (RStar reg)) env str
match (ROpt reg)        env []   = Just [[]]
match (ROpt reg)        env str  = match reg env str
match (RChar c)         env [c1] = if (c == c1) then Just [[]] else Nothing
match (RChar c)         env   _  = Nothing
match RDigit            env [c]  = if (isDigit c) then Just [[]] else Nothing
match RDigit            env   _  = Nothing
match RLetter           env [c]  = if (isAlpha c) then Just [[]] else Nothing
match RLetter           env   _  = Nothing
match RUpper            env [c]  = if (isUpper c) then Just [[]] else Nothing
match RUpper            env   _  = Nothing
match RLower            env [c]  = if (isLower c) then Just [[]] else Nothing
match RLower            env   _  = Nothing
match RAny              env [c]  = Just [[]]
match RAny              env   _  = Nothing
match (RStr s)          env s1   = if (s==s1) then Just [[]] else Nothing
match (RegVar id)       env str  = 
    case lookup id env of
     Just (Right reg) -> case matchRegexAll reg str of 
		  Just ([],_,[],_)  -> return $ [[(id,str)]]
		  _       -> Nothing                  
     Just (Left reg) -> case match reg env str of
                         Just _ ->  return $ [[(id,str)]]
                         _      -> Nothing
     Nothing  -> Just [[(id,str)]]    -- matches anything (?*)

to_regex :: Reg -> Either Reg Regex
to_regex r
  = case translate_to_regex r of
            Bad _ -> Left r
            Ok  rx -> Right rx

translate_to_regex :: Reg -> Err Regex
translate_to_regex reg = do s <- translate reg
                            return $ mkRegex s
  where translate r = case r of
         RAlt r r1 -> do s1 <- translate r 
                         s2 <- translate r1
                         return $ s1 ++ "|" ++ s2 
         RMinus r r1 -> Bad "minus"
         RSeq r r1 -> do s1 <- translate r
                         s2 <- translate r1
                         return $ s1 ++ s2
         RStar r   -> do s1 <- translate r 
                         return $ "(" ++ s1 ++ ")" ++ "*"
         RPlus r   -> do s1 <- translate r 
                         return $ "(" ++ s1 ++ ")" ++ "+"
         ROpt r    -> do s1 <- translate r 
                         return $ "(" ++ s1 ++ ")" ++ "?"
         REps      -> return $ " ?"
         RChar c   -> return $ [c]
         RAlts str -> return $ "[" ++ str ++ "]"
         RDigit -> return $ "[[:digit:]]"
         RLetter -> return $ "[[:alpha:]]"
         RUpper -> return $ "[[:upper:]]"
         RLower -> return $ "[[:lower:]]" 
         RAny   -> return $ "."
         RStr s -> return $ s
         RegVar _ -> Bad "variable found in translate_to_regex"

replaceVar :: Reg -> [(Ident, String)] -> Reg
replaceVar reg env =
    case reg of
     RAlt reg0 reg    -> RAlt (replaceVar reg0 env) (replaceVar reg env)
     RMinus reg0 reg  -> RMinus (replaceVar reg0 env) (replaceVar reg env)
     RSeq reg0 reg    -> RSeq (replaceVar reg0 env) (replaceVar reg env)
     RStar reg        -> RStar (replaceVar reg env)
     RPlus reg        -> RPlus (replaceVar reg env)
     ROpt reg         -> ROpt (replaceVar reg env)
     RegVar id        -> case lookup id env of
		          Nothing -> error "internal error in replaceVar"
		          Just s  -> (RStr s)
     x                -> x

recursiveDepth = 10

fixEnv :: [(Ident,Reg)] -> [(Ident,Reg)] -> Err [(Ident,Reg)]
fixEnv lenv env = fixEnv' lenv env 0
 where
  fixEnv' lenv env n
   | n == recursiveDepth = Bad $ "Maximum recursive depth: 10. Circular definitions in regexp?" 
   | otherwise = case [(id,reg) |(id,reg) <- lenv, containsVar reg] of
	           [] -> return lenv
	           _  -> do env' <- replace lenv env
			    fixEnv' env' env (n+1)  
  replace [] env = return []
  replace ((id,r):xs) env = do es <- replace xs env
			       r1 <- replaceEnv r env
			       return ((id,r1):es)

containsVar :: Reg  -> Bool
containsVar reg =
    case reg of
     RAlt reg0 reg    -> (containsVar reg0) || (containsVar reg )
     RMinus reg0 reg  -> (containsVar reg0) || (containsVar reg )
     RSeq reg0 reg    -> (containsVar reg0 ) || (containsVar reg )
     RStar reg        -> (containsVar reg )
     RPlus reg        -> (containsVar reg )
     ROpt reg         -> (containsVar reg)
     RegVar id        -> True
     x                -> False

replaceEnv :: Reg -> [(Ident,Reg)]  -> Err Reg
replaceEnv reg env =
    case reg of
     RAlt reg0 reg    -> do e1 <- replaceEnv reg0 env
			    e2 <- replaceEnv reg env
                            return $ RAlt e1 e2   
     RMinus reg0 reg  -> do e1 <- (replaceEnv reg0 env) 
			    e2 <- (replaceEnv reg env)
			    return $ RMinus e1 e2
     RSeq reg0 reg    -> do e1 <- (replaceEnv reg0 env) 
			    e2 <- (replaceEnv reg env) 
                            return $ RSeq e1 e2  
     RStar reg        -> do e1 <- (replaceEnv reg env) 
                            return $ RStar e1  
     RPlus reg        -> do e1 <- (replaceEnv reg env)
                            return $ RPlus e1 
     ROpt reg         -> do e1 <- (replaceEnv reg env)
			    return $ ROpt  e1 
     RegVar id@(Ident i) -> case lookup id env of
			     Nothing -> Bad $ "Unknown identifier regexp identifer: " ++ i
		             Just r  -> return $ r
     x                -> return x

split :: String -> [(String,String)] 
split s = zip (inits s) (tails s)

generateString :: Reg -> String
generateString reg = 
 case reg of
  RStr s         -> s 
  RSeq reg0 reg  -> generateString reg0 ++ generateString reg
  REps           -> []
  RChar c        -> [c]
  r              -> error $ "internal error in generateString" ++ show r

collect_var_reg :: Reg -> [Ident]
collect_var_reg reg = 
 case reg of
  RStr _         -> []
  RSeq reg0 reg  -> collect_var_reg reg0 ++ collect_var_reg reg
  RegVar id      -> [id]
  REps           -> []
  RChar _        -> []
  r              -> error $ "internal error in collect_var_reg" ++ show r
