module Paradigm where

import IO ( stdin, hGetContents )
import System ( getArgs, getProgName )

import LexExtract
import ParExtract
import SkelExtract
import PrintExtract
import AbsExtract
import MatchReg
import PrintExtract
import List (nub, (\\))
import Text.Regex
import UTF8
import ErrM

type Name = String

emptyPName = ""

type Paradigm  = (Name, Int , Arguments, Constraints, LocalRegEnv)

name :: Paradigm -> Name
name (n,_,_,_,_) = n

p_id :: Paradigm -> String
p_id (n,i,_,_,_) = show i 

args :: Paradigm -> Arguments
args  (_,_,as,_,_) = as

constraints :: Paradigm -> RegLogic
constraints (_,_,_,(l,_),_) = l

localenv  :: Paradigm -> LocalRegEnv
localenv (_,_,_,_,env) = env

type Constraints = (RegLogic,[Ident])

type Arguments   = [Reg]

type Paradigms   = [Paradigm]

type RegEnv      = [(Ident, Reg)]

type LocalRegEnv = [(Ident, Either Reg Regex)]

data RegLogic =
   RConj RegLogic RegLogic
 | RDisj RegLogic RegLogic
 | RNeg RegLogic
 | RL Reg
  deriving (Eq,Ord,Show)

items_to_reg :: [Item] -> RegEnv -> [Binding] -> Err (Reg,LocalRegEnv)
items_to_reg xs env bs = do reg <- convert_it xs
                            ls  <- convert_local env bs
                            return (reg, ls)
    where
     convert_it [x]    = transform_item x
     convert_it (x:xs) = 
         do r1 <- transform_item x
	    r2 <- convert_it xs
	    return $ RSeq r1 r2
     transform_item i  =
         case i of
         (Var id)       -> return $ RegVar id
         (C s)          -> return $ RStr s


convert_local :: RegEnv -> [Binding] -> Err LocalRegEnv
convert_local rs bs = do ls <- fixEnv [(i,r) | Bind i r <- bs] rs
                         return $ [(i,to_regex r) | (i,r) <- ls]

transform_logic :: Logic -> Err RegLogic
transform_logic l =
 case l of
  Conj logic0 logic  -> do l1 <- transform_logic logic0
			   l2 <- transform_logic logic
                           return (RConj l1 l2)
  Disj logic0 logic  -> do l1 <- transform_logic logic0
			   l2 <- transform_logic logic
			   return (RDisj l1 l2) 
  Neg logic          -> do l1 <- transform_logic logic
			   return (RNeg l1)
  LPatt (P pattern)  -> case items_to_reg pattern [] [] of
			 Ok (reg,[]) -> return (RL reg)
			 _           -> Bad $ "Regular expressions may not appear in the constraints" 

get_paradigms :: FilePath -> Bool -> IO (Err Paradigms)
get_paradigms f utf8
                = do s <- readFile f
		     case pGrammar (convert (myLexer s)) of
		      Bad s   -> return (Bad s)
		      Ok tree -> return $ collect_paradigms tree
 where convert xs = if utf8 then map conv xs else xs
       conv (PT p t) = PT p (conv' t)
       conv x        = x
       conv'  (TS s) = TS $ decodeUTF8 s
       conv'  (TL s) = TL $ decodeUTF8 s
       conv'  (TI s) = TI $ decodeUTF8 s
       conv'  (TV s) = TV $ decodeUTF8 s
       conv'  (TD s) = TD $ decodeUTF8 s
       conv'  (TC s) = TC $ decodeUTF8 s

collect_paradigms (Grammar defs) = 
   do let xs = [ (i,r) | (Regdef i r) <- defs]
      env <- fixEnv xs xs
      let para1 = [ (i,[items_to_reg pt env localenv | (P pt) <- xs], transform_logic l) | (Paradigm (Ident i) (E localenv) xs l) <- defs]
      let para2 = [ (i,[items_to_reg pt env [] | (P pt) <- xs], transform_logic l) | (ParadigmN (Ident i) xs l) <- defs]
      typecheck $ (para1 ++ para2)
  
positive_match :: Paradigm -> String -> [[(Ident, String)]]
positive_match (_,_,_,(l,xs),env) s = pmatch l
 where pmatch l =
        case l of
	 RConj logic0 logic -> pmatch logic0 ++ pmatch logic 
	 RDisj logic0 logic -> pmatch logic0 ++ pmatch logic 
	 RNeg  logic        -> pmatch logic
	 RL    reg          -> 
	     case (xs \\ (collect_var_reg reg)) of
	      [] -> maybe [] id (match reg env s) 
	      _  -> []

collect_vars :: RegLogic -> [Ident]
collect_vars l = nub $ collect l
  where collect lgc =
         case lgc of
	  RConj logic0 logic -> collect logic0 ++ collect logic 
	  RDisj logic0 logic -> collect logic0 ++ collect logic 
	  RNeg  logic        -> collect logic
	  RL    reg          -> collect_var_reg reg

typecheck ps = case all_paradigms ps of
	        Ok paras -> validate_properties [ (n,i,a,c,l) | (i,(n,a,c,l)) <- zip [1..] paras]
		Bad x    -> Bad x
 where
  all_paradigms [] = return [] 
  all_paradigms (x:xs) =  do p    <- one_paradigm x
			     rest <- all_paradigms xs
			     return (p:rest)
  one_paradigm (i, xs, l) = 
      case ([x | (Bad x) <- xs],l) of
       ([], Ok l) -> return (i, [x | (Ok (x,_)) <- xs], (l, collect_vars l), 
			        concat [ys | (Ok (_,ys)) <- xs])
       ([], Bad s) -> Bad $ "Error(s) in paradigm " ++ i ++ ":\n" ++ ("  " ++ s)
       (xs, Bad s) -> Bad $ "Error(s) in paradigm " ++ i ++ ":\n" ++ unlines (map ("  " ++) (s:xs))
       (xs, _)     -> Bad $ "Error(s) in paradigm " ++ i ++ ":\n" ++ unlines (map ("  " ++) xs)

validate_properties :: Paradigms -> Err Paradigms
validate_properties = return . map validate
   where validate = id
