module Dictionary (Dict(..),
		   Dictionary,
		   Entry,
		   prTable,
		   removeAttr,
		   FullFormLex,
		   Ident,
		   classifyDict,
		   noAttr,
		   entry,
		   entryI,
	           entryWithInfo,
                   entryWithInfoI,
		   EntryN,
		   dictionary,
		   unDict,
		   size,
		   sizeW,
		   unionDictionary,
		   unionDictionaries,
                   emptyDict,
		   dict2fullform,
		   buildTrie,
                   shareDictionary
		  ) where

import General
import List (sortBy, group)
import Char
import FMTrie
import qualified Data.Set as Set
import SharedString

-- untyped dictionary: dictionary word, category, inherent features, inflection

class Param a => Dict a where
  dictword :: (a -> Str) -> String
  dictword f = concat $ take 1 $ unStr (f value0)
  category :: (a -> Str) -> String
  category = const "Undefined"
  defaultAttr :: (a -> Str) -> Attr
  defaultAttr = const noComp
  attrException :: (a -> Str) -> [(a,Attr)]
  attrException = const []
  
data Dictionary       = Dict [Entry]
type Dictionary_Word  = String
type Category         = String
type Inherent         = String
type Untyped          = String
type Inflection_Table = [(Untyped,(Attr,Str))]
type Extra            = [Str] -- 

-- Str should be polymorphic, parametric printing
-- Especially if it should be able to describe phonetics.
-- Paradigm Name should be a part of the Dictionary.

type Entry = (Dictionary_Word, Category, [Inherent], Inflection_Table,Extra)

type Ident = String

type EntryN =  (Dictionary_Word, Category, [Inherent], [(Ident,Str)])
-- Should have the extra information?

emptyDict :: Dictionary
emptyDict = Dict []

shareDictionary :: Dictionary -> Dictionary
shareDictionary (Dict xs) = Dict $ map shareEntry xs

shareEntry :: Entry -> Entry
shareEntry (dict_word, cat, inhs, infl, extra) =
    (shareString dict_word, shareString cat, map shareString inhs, shareTable infl, extra) 
 where shareTable xs = map (\(a,(p,b)) -> (shareString a,(p,shareStr b))) xs

infTable :: Dict a => (a -> Str) -> Inflection_Table
infTable f = prTableAttr f (defaultAttr f) (attrException f)

extraTable ::  Dict a => (a -> Str) -> Extra
extraTable f = [s | (_,s) <- table f]
  
entry  :: Dict a => (a -> Str) -> Entry
entry f = entryI f []
  
entryI :: Dict a => (a -> Str) -> [Inherent] -> Entry
entryI f ihs = (dictword f, category f, ihs, infTable f,[])

entryWithInfo ::  Dict a => (a -> (Str,Str)) -> Entry
entryWithInfo f = entryWithInfoI f []

entryWithInfoI ::  Dict a => (a -> (Str,Str)) -> [Inherent] -> Entry
entryWithInfoI fun ihs = (dictword f, category f, ihs, infTable f,extraTable g)
  where f = \a -> fst (fun a)
        g = \a -> snd (fun a)
-- entryI (\a -> unionStr (f a) (unionStr (mkStr "|") (g a))) ihs

prTableAttr :: Param a => (a -> Str) -> Attr -> [(a,Attr)] -> [(String,(Attr,Str))]
prTableAttr t da ts = 
    [(prValue a,(maybe da id (lookup a ts),s)) | (a,s) <- table t]
 
prTableW :: Param a => Table a -> [(String,(Attr,Str))]
prTableW t = [ (a,(noComp,s)) | (a,s) <- prTable t]

prTable :: Param a => Table a -> Table String
prTable = map (\ (a,b) -> (prValue a, b))

unDict :: Dictionary -> [Entry]
unDict (Dict xs) = xs

size :: Dictionary -> Int
size = length . unDict

sizeW :: Dictionary -> Int
sizeW dict = sum [length t | (_,_,_,t,_) <- unDict dict]

sizeEntry :: Entry -> Int
sizeEntry (_,_,_,t,_) = length t

dictionary :: [Entry] -> Dictionary
dictionary = Dict

unionDictionary :: Dictionary -> Dictionary -> Dictionary
unionDictionary (Dict xs) (Dict ys) = Dict $ xs ++ ys

unionDictionaries :: [Dictionary] -> Dictionary
unionDictionaries = foldr unionDictionary emptyDict

removeAttr :: Dictionary -> [EntryN]
removeAttr = map noAttr . unDict


-- Ignores the extra information! 
noAttr :: Entry ->  EntryN
noAttr (d,c,inh,tab,_) = (d,c,inh,[(i, s) | (i,(_,s)) <- tab])

-- group a dictionary into categories; reverses the entries... 
classifyDict :: Dictionary -> [(Ident,[Entry])]
classifyDict = foldr addNext [] . unDict
 where
  addNext entry@(_,cat,_,_,_) dict = case dict of
    (c,es) : dict' | cat == c -> (c, entry:es) : dict'
    ces    : dict'            -> ces           : addNext entry dict'
    []                        -> [(cat,[entry])]

-- full-form lexicon: show all different analyses of a word as strings

type FullFormLex = [(String,[(Attr,String)])]

dict2fullform :: Dictionary -> FullFormLex
dict2fullform (Dict es) = f es Set.empty 
  where f [] _ = []
        f (e:es) set = case entry2full e set of
                         (xss,new_set) -> 
                             [(s,[(a,show d) | (a,d) <- xs]) | (s,xs) <- xss] ++ f es new_set

entry2full :: Entry -> Set.Set String -> ([(String,[(Attr,LexData)])],Set.Set String)
entry2full (stem, typ, inhs, infl,extra) set = 
  (concatMap mkForm (zip infl (extra ++ repeat nonExist)),new_set) where
    (name,new_set) = get_name stem typ set
    mkForm ((par,(a,str)),extra) = [(s1, [(a, LD stem (name ++":"++w_id) typ par inhs extra)]) | 
                                          (s1,w_id,extra) <- (zip3
                                                         (unStr str)
                                                         (map show [1..])
                                                         (unStr extra ++ (repeat [])))]

get_name :: String -> String -> Set.Set String -> (String,Set.Set String)
get_name stem typ set = f 0
 where
  st = stem++"_"++typ
  f 0 
   | Set.member st set = f 2
   | otherwise         = ((nospace st), Set.insert st set) 
  f n = case stem ++ show n ++"_"++ typ of
         st' -> if (Set.member st' set) then
                    f (n+1)
                 else ((nospace st'), Set.insert st set) 
  nospace = map (\c -> if (c==' ') then '_' else c)

{-
dict2fullform :: Dictionary -> FullFormLex
dict2fullform d =  [ (s,[(a,show d) | (a,d) <- xs]) | (s,xs) <- concatMap entry2full $ zip (unDict d) [0..]]
  
entry2full :: (Entry,Int) -> [(String,[(Attr,LexData)])]
entry2full ((stem, typ, inhs, infl,extra),n) = 
  concatMap mkForm (zip infl (extra ++ repeat nonExist)) where
    mkForm ((par,(a,str)),extra) = [(s1, [(a, LD stem (show n ++":"++w_id) typ par inhs extra)]) | 
                                          (s1,w_id,extra) <- (zip3
                                                         (unStr str)
                                                         (map show [1..])
                                                         (unStr extra ++ (repeat [])))]
-}

buildTrie :: Dictionary -> SATrie
buildTrie = tcompile . dict2fullformwithLexData 
 where dict2fullformwithLexData (Dict es) = f es Set.empty 
       f [] _ = []
       f (e:es) set = case entry2full e set of
                       (xss,new_set) -> 
                           xss ++ f es new_set

-- concatMap entry2full $ zip (unDict dict) [0..]

-- Adapted merge sort from List.hs 
sortAssocs :: [(String,(Attr,LexData))] -> [(String,[(Attr,LexData)])]
sortAssocs = mergesort (\(a,_) -> \(b,_) -> compare a b)
  where
    mergesort cmp = mergesort' cmp . map wrap
    mergesort' cmp [] = []
    mergesort' cmp [xs] = xs
    mergesort' cmp xss = mergesort' cmp (merge_pairs cmp xss)

    merge_pairs cmp [] = []
    merge_pairs cmp [xs] = [xs]
    merge_pairs cmp (xs:ys:xss) = merge cmp xs ys : merge_pairs cmp xss
    wrap (a,b) = [(a,[b])]
    merge cmp xs [] = xs
    merge cmp [] ys = ys
    merge cmp (x@(a,xs):xss) (y@(_,ys):yss)
     = case x `cmp` y of
        GT -> y : merge cmp (x:xss)   yss
	EQ -> case xs of
               [z] -> merge cmp ((a,z:ys):xss) yss 
	       zs  -> merge cmp ((a,zs++ys):xss) yss 
        _  -> x : merge cmp    xss (y:yss)
